@@ -33,7 +33,8 @@
 #include <gtksourceview/gtksourceview.h>
 #include <gtksourceview/gtksourcelanguage.h>
 #include <gtksourceview/gtksourcelanguagemanager.h>
-#include <gtksourceview/gtksourcestylemanager.h>
+#include <gtksourceview/gtksourcestyleschememanager.h>
+#include <gtksourceview/gtksourceprintcompositor.h>
 #ifdef TEST_XML_MEM
 #include <libxml/xmlreader.h>
 #endif
@@ -52,14 +53,16 @@ static GtkSourceStyleScheme *style_scheme = NULL;
 
 #define READ_BUFFER_SIZE   4096
 
-#define MARKER_TYPE_1      "one"
-#define MARKER_TYPE_2      "two"
+#define MARK_TYPE_1      "one"
+#define MARK_TYPE_2      "two"
 
 
 /* Private prototypes -------------------------------------------------------- */
 
 static void       open_file_cb                   (GtkAction       *action,
 						  gpointer         user_data);
+static void       print_file_cb                  (GtkAction       *action,
+						  gpointer         user_data);						  
 static void       debug_thing_3_cb		 (GtkAction       *action,
 						  gpointer         user_data);
 
@@ -67,12 +70,14 @@ static void       new_view_cb                    (GtkAction       *action,
 						  gpointer         user_data);
 static void       numbers_toggled_cb             (GtkAction       *action,
 						  gpointer         user_data);
-static void       markers_toggled_cb             (GtkAction       *action,
+static void       marks_toggled_cb               (GtkAction       *action,
 						  gpointer         user_data);
 static void       margin_toggled_cb              (GtkAction       *action,
 						  gpointer         user_data);
 static void       hl_line_toggled_cb             (GtkAction       *action,
 						  gpointer         user_data);
+static void       wrap_lines_toggled_cb          (GtkAction       *action,
+						  gpointer         user_data);
 static void       auto_indent_toggled_cb         (GtkAction       *action,
 						  gpointer         user_data);
 static void       insert_spaces_toggled_cb       (GtkAction       *action,
@@ -80,6 +85,9 @@ static void       insert_spaces_toggled_cb       (GtkAction       *action,
 static void       tabs_toggled_cb                (GtkAction       *action,
 						  GtkAction       *current,
 						  gpointer         user_data);
+static void       indent_toggled_cb              (GtkAction       *action,
+						  GtkAction       *current,
+						  gpointer         user_data);
 
 static GtkWidget *create_view_window             (GtkSourceBuffer *buffer,
 						  GtkSourceView   *from);
@@ -96,10 +104,13 @@ static GtkActionEntry buffer_action_entries[] = {
 
 static GtkActionEntry view_action_entries[] = {
 	{ "FileMenu", NULL, "_File" },
+	{ "Print", GTK_STOCK_PRINT, "_Print", "<control>P",
+	  "Print the current file", G_CALLBACK (print_file_cb) },
 	{ "ViewMenu", NULL, "_View" },
 	{ "NewView", GTK_STOCK_NEW, "_New View", NULL,
 	  "Create a new view of the file", G_CALLBACK (new_view_cb) },
-	{ "TabsWidth", NULL, "_Tabs Width" },
+	{ "TabWidth", NULL, "_Tab Width" },
+	{ "IndentWidth", NULL, "I_ndent Width" },
 	{ "SmartHomeEnd", NULL, "_Smart Home/End" },
 	{ "DebugThing3", GTK_STOCK_FIND_AND_REPLACE, "Search and _Replace", "<control>R",
 	  "Search and Replace", G_CALLBACK (debug_thing_3_cb) },
@@ -109,15 +120,18 @@ static GtkToggleActionEntry toggle_entries[] = {
 	{ "ShowNumbers", NULL, "Show _Line Numbers", NULL,
 	  "Toggle visibility of line numbers in the left margin",
 	  G_CALLBACK (numbers_toggled_cb), FALSE },
-	{ "ShowMarkers", NULL, "Show _Markers", NULL,
-	  "Toggle visibility of markers in the left margin",
-	  G_CALLBACK (markers_toggled_cb), FALSE },
-	{ "ShowMargin", NULL, "Show M_argin", NULL,
+	{ "ShowMarks", NULL, "Show Line _Marks", NULL,
+	  "Toggle visibility of marks in the left margin",
+	  G_CALLBACK (marks_toggled_cb), FALSE },
+	{ "ShowMargin", NULL, "Show Right M_argin", NULL,
 	  "Toggle visibility of right margin indicator",
 	  G_CALLBACK (margin_toggled_cb), FALSE },
 	{ "HlLine", NULL, "_Highlight Current Line", NULL,
 	  "Toggle highlighting of current line",
 	  G_CALLBACK (hl_line_toggled_cb), FALSE },
+	{ "WrapLines", NULL, "_Wrap Lines", NULL,
+	  "Toggle line wrapping",
+	  G_CALLBACK (wrap_lines_toggled_cb), FALSE },
 	{ "AutoIndent", NULL, "Enable _Auto Indent", NULL,
 	  "Toggle automatic auto indentation of text",
 	  G_CALLBACK (auto_indent_toggled_cb), FALSE },
@@ -126,12 +140,21 @@ static GtkToggleActionEntry toggle_entries[] = {
 	  G_CALLBACK (insert_spaces_toggled_cb), FALSE }
 };
 
-static GtkRadioActionEntry radio_entries[] = {
-	{ "TabsWidth4", NULL, "4", NULL, "Set tabulation width to 4 spaces", 4 },
-	{ "TabsWidth6", NULL, "6", NULL, "Set tabulation width to 6 spaces", 6 },
-	{ "TabsWidth8", NULL, "8", NULL, "Set tabulation width to 8 spaces", 8 },
-	{ "TabsWidth10", NULL, "10", NULL, "Set tabulation width to 10 spaces", 10 },
-	{ "TabsWidth12", NULL, "12", NULL, "Set tabulation width to 12 spaces", 12 }
+static GtkRadioActionEntry tabs_radio_entries[] = {
+	{ "TabWidth4", NULL, "4", NULL, "Set tabulation width to 4 spaces", 4 },
+	{ "TabWidth6", NULL, "6", NULL, "Set tabulation width to 6 spaces", 6 },
+	{ "TabWidth8", NULL, "8", NULL, "Set tabulation width to 8 spaces", 8 },
+	{ "TabWidth10", NULL, "10", NULL, "Set tabulation width to 10 spaces", 10 },
+	{ "TabWidth12", NULL, "12", NULL, "Set tabulation width to 12 spaces", 12 }
+};
+
+static GtkRadioActionEntry indent_radio_entries[] = {
+	{ "IndentWidthUnset", NULL, "Use Tab Width", NULL, "Set indent width same as tab width", -1 },
+	{ "IndentWidth4", NULL, "4", NULL, "Set indent width to 4 spaces", 4 },
+	{ "IndentWidth6", NULL, "6", NULL, "Set indent width to 6 spaces", 6 },
+	{ "IndentWidth8", NULL, "8", NULL, "Set indent width to 8 spaces", 8 },
+	{ "IndentWidth10", NULL, "10", NULL, "Set indent width to 10 spaces", 10 },
+	{ "IndentWidth12", NULL, "12", NULL, "Set indent width to 12 spaces", 12 }
 };
 
 static GtkRadioActionEntry smart_home_end_entries[] = {
@@ -157,19 +180,28 @@ static const gchar *view_ui_description =
 "      <menuitem action=\"NewView\"/>"
 "      <separator/>"
 "      <menuitem action=\"ShowNumbers\"/>"
-"      <menuitem action=\"ShowMarkers\"/>"
+"      <menuitem action=\"ShowMarks\"/>"
 "      <menuitem action=\"ShowMargin\"/>"
 "      <menuitem action=\"HlLine\"/>"
+"      <menuitem action=\"WrapLines\"/>"
 "      <separator/>"
 "      <menuitem action=\"AutoIndent\"/>"
 "      <menuitem action=\"InsertSpaces\"/>"
 "      <separator/>"
-"      <menu action=\"TabsWidth\">"
-"        <menuitem action=\"TabsWidth4\"/>"
-"        <menuitem action=\"TabsWidth6\"/>"
-"        <menuitem action=\"TabsWidth8\"/>"
-"        <menuitem action=\"TabsWidth10\"/>"
-"        <menuitem action=\"TabsWidth12\"/>"
+"      <menu action=\"TabWidth\">"
+"        <menuitem action=\"TabWidth4\"/>"
+"        <menuitem action=\"TabWidth6\"/>"
+"        <menuitem action=\"TabWidth8\"/>"
+"        <menuitem action=\"TabWidth10\"/>"
+"        <menuitem action=\"TabWidth12\"/>"
+"      </menu>"
+"      <menu action=\"IndentWidth\">"
+"        <menuitem action=\"IndentWidthUnset\"/>"
+"        <menuitem action=\"IndentWidth4\"/>"
+"        <menuitem action=\"IndentWidth6\"/>"
+"        <menuitem action=\"IndentWidth8\"/>"
+"        <menuitem action=\"IndentWidth10\"/>"
+"        <menuitem action=\"IndentWidth12\"/>"
 "      </menu>"
 "      <separator/>"
 "      <menu action=\"SmartHomeEnd\">"
@@ -187,6 +219,7 @@ static const gchar *buffer_ui_description =
 "  <menubar name=\"MainMenu\">"
 "    <menu action=\"FileMenu\">"
 "      <menuitem action=\"Open\"/>"
+"      <menuitem action=\"Print\"/>"
 "      <menuitem action=\"DebugThing3\"/>"
 "      <separator/>"
 "      <menuitem action=\"Quit\"/>"
@@ -263,20 +296,13 @@ gtk_source_buffer_load_file (GtkSourceBuffer *source_buffer,
 }
 
 static void
-remove_all_markers (GtkSourceBuffer *buffer)
+remove_all_marks (GtkSourceBuffer *buffer)
 {
-	GSList *markers;
-	GtkTextIter begin, end;
+	GtkTextIter s, e;
 
-	gtk_text_buffer_get_bounds (GTK_TEXT_BUFFER (buffer), &begin, &end);
-	markers = gtk_source_buffer_get_markers_in_region (buffer, &begin, &end);
-	while (markers)
-	{
-		GtkSourceMarker *marker = markers->data;
+	gtk_text_buffer_get_bounds (GTK_TEXT_BUFFER (buffer), &s, &e);
 
-		gtk_source_buffer_delete_marker (buffer, marker);
-		markers = g_slist_delete_link (markers, markers);
-	}
+	gtk_source_buffer_remove_marks (buffer, &s, &e, NULL);
 }
 
 /* Note this is wrong for several reasons, e.g. g_pattern_match is broken
@@ -284,7 +310,7 @@ remove_all_markers (GtkSourceBuffer *buffer)
 static GtkSourceLanguage *
 get_language_for_filename (const gchar *filename)
 {
-	GSList *list;
+	const gchar * const *languages;
 	gchar *filename_utf8;
 	GtkSourceLanguageManager *manager;
 
@@ -292,15 +318,17 @@ get_language_for_filename (const gchar *filename)
 	g_return_val_if_fail (filename_utf8 != NULL, NULL);
 
 	manager = gtk_source_language_manager_get_default ();
-	list = gtk_source_language_manager_list_languages (manager);
+	languages = gtk_source_language_manager_get_language_ids (manager);
 
-	while (list != NULL)
+	while (*languages != NULL)
 	{
 		GtkSourceLanguage *lang;
 		gchar **globs, **p;
 
-		lang = list->data;
-		list = g_slist_delete_link (list, list);
+		lang = gtk_source_language_manager_get_language (manager,
+								 *languages);
+		g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE (lang), NULL);
+		++languages;
 
 		globs = gtk_source_language_get_globs (lang);
 		if (globs == NULL)
@@ -312,7 +340,7 @@ get_language_for_filename (const gchar *filename)
 			{
 				g_strfreev (globs);
 				g_free (filename_utf8);
-				g_slist_free (list);
+
 				return lang;
 			}
 		}
@@ -330,19 +358,21 @@ get_language_for_filename (const gchar *filename)
 static GtkSourceLanguage *
 get_language_for_mime_type (const gchar *mime)
 {
-	GSList *list;
+	const gchar * const *languages;
 	GtkSourceLanguageManager *manager;
 
 	manager = gtk_source_language_manager_get_default ();
-	list = gtk_source_language_manager_list_languages (manager);
+	languages = gtk_source_language_manager_get_language_ids (manager);
 
-	while (list != NULL)
+	while (*languages != NULL)
 	{
 		GtkSourceLanguage *lang;
 		gchar **mimetypes, **p;
 
-		lang = list->data;
-		list = g_slist_delete_link (list, list);
+		lang = gtk_source_language_manager_get_language (manager,
+								 *languages);
+		g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE (lang), NULL);
+		++languages;
 
 		mimetypes = gtk_source_language_get_mime_types (lang);
 
@@ -354,7 +384,6 @@ get_language_for_mime_type (const gchar *mime)
 			if (strcmp (*p, mime) == 0)
 			{
 				g_strfreev (mimetypes);
-				g_slist_free (list);
 				return lang;
 			}
 		}
@@ -402,6 +431,51 @@ get_language_for_file (const gchar *filename)
 	return language;
 }
 
+static GtkSourceLanguage *
+get_language_by_id (const gchar *id)
+{
+	GtkSourceLanguageManager *manager;
+	manager = gtk_source_language_manager_get_default ();
+	return gtk_source_language_manager_get_language (manager, id);
+}
+
+static GtkSourceLanguage *
+get_language (GtkTextBuffer *buffer, const gchar *filename)
+{
+	GtkSourceLanguage *language = NULL;
+	GtkTextIter start, end;
+	gchar *text;
+	const gchar *lang_string;
+
+	gtk_text_buffer_get_start_iter (buffer, &start);
+	end = start;
+	gtk_text_iter_forward_line (&end);
+
+#define LANG_STRING "gtk-source-lang:"
+	text = gtk_text_iter_get_slice (&start, &end);
+	lang_string = strstr (text, LANG_STRING);
+	if (lang_string != NULL)
+	{
+		gchar **tokens;
+
+		lang_string += strlen (LANG_STRING);
+		g_strchug (lang_string);
+
+		tokens = g_strsplit_set (lang_string, " \t\n", 2);
+
+		if (tokens != NULL && tokens[0] != NULL)
+			language = get_language_by_id (tokens[0]);
+
+		g_strfreev (tokens);
+	}
+
+	if (!language)
+		language = get_language_for_file (filename);
+
+	g_free (text);
+	return language;
+}
+
 static gboolean
 open_file (GtkSourceBuffer *buffer, const gchar *filename)
 {
@@ -417,14 +491,14 @@ open_file (GtkSourceBuffer *buffer, const gchar *filename)
 		g_free (curdir);
 	}
 
-	remove_all_markers (buffer);
+	remove_all_marks (buffer);
 
 	success = gtk_source_buffer_load_file (buffer, filename, NULL);
 
 	if (!success)
 		goto out;
 
-	language = get_language_for_file (filename);
+	language = get_language (GTK_TEXT_BUFFER (buffer), filename);
 
 	if (language == NULL)
 		g_print ("No language found for file `%s'\n", filename);
@@ -434,7 +508,37 @@ open_file (GtkSourceBuffer *buffer, const gchar *filename)
 				"filename", g_strdup (filename),
 				(GDestroyNotify) g_free);
 
+	if (language != NULL)
+	{
+		gchar **styles;
 
+		styles = gtk_source_language_get_style_ids (language);
+
+		if (styles == NULL)
+			g_print ("No styles in language '%s'\n", gtk_source_language_get_name (language));
+		else
+		{
+			gchar **ids;
+			g_print ("Styles in in language '%s':\n", gtk_source_language_get_name (language));
+
+			ids = styles;
+
+			while (*ids != NULL)
+			{
+				const gchar *name;
+
+				name = gtk_source_language_get_style_name (language, *ids);
+
+				g_print ("- %s (name: '%s')\n", *ids, name);
+
+				++ids;
+			}
+
+			g_strfreev (styles);
+		}
+
+		g_print("\n");
+	}
 out:
 	g_free (freeme);
 	return success;
@@ -453,10 +557,10 @@ numbers_toggled_cb (GtkAction *action, gpointer user_data)
 }
 
 static void
-markers_toggled_cb (GtkAction *action, gpointer user_data)
+marks_toggled_cb (GtkAction *action, gpointer user_data)
 {
 	g_return_if_fail (GTK_IS_TOGGLE_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
-	gtk_source_view_set_show_line_markers (
+	gtk_source_view_set_show_line_marks (
 		GTK_SOURCE_VIEW (user_data),
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
@@ -465,7 +569,7 @@ static void
 margin_toggled_cb (GtkAction *action, gpointer user_data)
 {
 	g_return_if_fail (GTK_IS_TOGGLE_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
-	gtk_source_view_set_show_margin (
+	gtk_source_view_set_show_right_margin (
 		GTK_SOURCE_VIEW (user_data),
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
@@ -479,6 +583,14 @@ hl_line_toggled_cb (GtkAction *action, gpointer user_data)
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
 
+static void
+wrap_lines_toggled_cb (GtkAction *action, gpointer user_data)
+{
+	g_return_if_fail (GTK_IS_TOGGLE_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
+	gtk_text_view_set_wrap_mode (user_data,
+				     gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)) ?
+					GTK_WRAP_WORD : GTK_WRAP_NONE);
+}
 
 static void
 auto_indent_toggled_cb (GtkAction *action,
@@ -506,7 +618,18 @@ tabs_toggled_cb (GtkAction *action,
 		 gpointer user_data)
 {
 	g_return_if_fail (GTK_IS_RADIO_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
-	gtk_source_view_set_tabs_width (
+	gtk_source_view_set_tab_width (
+		GTK_SOURCE_VIEW (user_data),
+		gtk_radio_action_get_current_value (GTK_RADIO_ACTION (action)));
+}
+
+static void
+indent_toggled_cb (GtkAction *action,
+		   GtkAction *current,
+		   gpointer user_data)
+{
+	g_return_if_fail (GTK_IS_RADIO_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
+	gtk_source_view_set_indent_width (
 		GTK_SOURCE_VIEW (user_data),
 		gtk_radio_action_get_current_value (GTK_RADIO_ACTION (action)));
 }
@@ -677,6 +800,216 @@ open_file_cb (GtkAction *action, gpointer user_data)
 	gtk_widget_destroy (chooser);
 }
 
+#define NON_BLOCKING_PAGINATION
+
+#ifndef NON_BLOCKING_PAGINATION
+
+static void
+begin_print (GtkPrintOperation        *operation, 
+	     GtkPrintContext          *context,
+	     GtkSourcePrintCompositor *compositor)
+{
+	gint n_pages;
+	
+	while (!gtk_source_print_compositor_paginate (compositor, context))
+		;
+
+	n_pages = gtk_source_print_compositor_get_n_pages (compositor);
+	gtk_print_operation_set_n_pages (operation, n_pages);
+}
+
+#else
+
+static gboolean
+paginate (GtkPrintOperation        *operation, 
+	  GtkPrintContext          *context,
+	  GtkSourcePrintCompositor *compositor)
+{
+	g_print ("Pagination progress: %.2f %%\n", gtk_source_print_compositor_get_pagination_progress (compositor) * 100.0);
+	
+	if (gtk_source_print_compositor_paginate (compositor, context))
+	{
+		gint n_pages;
+
+		g_assert (gtk_source_print_compositor_get_pagination_progress (compositor) == 1.0);
+		g_print ("Pagination progress: %.2f %%\n", gtk_source_print_compositor_get_pagination_progress (compositor) * 100.0);
+		        
+		n_pages = gtk_source_print_compositor_get_n_pages (compositor);
+		gtk_print_operation_set_n_pages (operation, n_pages);
+
+
+		
+		return TRUE;
+	}
+     
+	return FALSE;
+}
+
+#endif
+
+#define ENABLE_CUSTOM_OVERLAY
+
+static void
+draw_page (GtkPrintOperation        *operation,
+	   GtkPrintContext          *context,
+	   gint                      page_nr,
+	   GtkSourcePrintCompositor *compositor)
+{
+#ifdef ENABLE_CUSTOM_OVERLAY
+
+	/* This part of the code shows how to add a custom overlay to the
+	   printed text generated by GtkSourcePrintCompositor */
+
+	cairo_t *cr;
+	PangoLayout *layout;
+	PangoFontDescription *desc;
+	PangoRectangle rect;
+
+	
+	cr = gtk_print_context_get_cairo_context (context);
+	
+	cairo_save (cr);
+	
+	layout = gtk_print_context_create_pango_layout (context);
+
+	pango_layout_set_text (layout, "Draft", -1);
+
+	desc = pango_font_description_from_string ("Sans Bold 120");
+	pango_layout_set_font_description (layout, desc);
+	pango_font_description_free (desc);
+
+	
+	pango_layout_get_extents (layout, NULL, &rect);
+		
+  	cairo_move_to (cr, 
+  		       (gtk_print_context_get_width (context) - ((double) rect.width / (double) PANGO_SCALE)) / 2,
+  		       (gtk_print_context_get_height (context) - ((double) rect.height / (double) PANGO_SCALE)) / 2);
+	  		       
+	pango_cairo_layout_path (cr, layout);
+	
+  	/* Font Outline */
+	cairo_set_source_rgba (cr, 0.85, 0.85, 0.85, 0.80);
+	cairo_set_line_width (cr, 0.5);
+	cairo_stroke_preserve (cr);
+
+	/* Font Fill */
+	cairo_set_source_rgba (cr, 0.8, 0.8, 0.8, 0.60);
+	cairo_fill (cr);
+	  	  
+	g_object_unref (layout);
+	cairo_restore (cr);
+#endif
+
+	/* To print page_nr you only need to call the following function */
+	gtk_source_print_compositor_draw_page (compositor, context, page_nr);
+}
+
+static void
+end_print (GtkPrintOperation        *operation, 
+	   GtkPrintContext          *context,
+	   GtkSourcePrintCompositor *compositor)
+{
+	g_object_unref (compositor);
+}
+
+#define LINE_NUMBERS_FONT_NAME	"Sans 8"
+#define HEADER_FONT_NAME	"Sans 11"
+#define FOOTER_FONT_NAME	"Sans 11"
+#define BODY_FONT_NAME		"Monospace 9"
+	
+/*
+#define SETUP_FROM_VIEW 
+*/
+
+#undef SETUP_FROM_VIEW 
+
+
+static void
+print_file_cb (GtkAction *action, gpointer user_data)
+{
+	GtkSourceView *view;
+	GtkSourceBuffer *buffer;
+	GtkSourcePrintCompositor *compositor;
+	GtkPrintOperation *operation;	
+	const gchar *filename;
+	gchar *basename;
+	
+	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
+	
+	view = GTK_SOURCE_VIEW (user_data);
+
+	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view))); 
+	
+	filename = g_object_get_data (G_OBJECT (buffer), "filename");	
+	basename = g_filename_display_basename (filename);
+	
+#ifdef SETUP_FROM_VIEW
+	compositor = gtk_source_print_compositor_new_from_view (view);
+#else
+	
+	compositor = gtk_source_print_compositor_new (buffer);
+
+	gtk_source_print_compositor_set_tab_width (compositor,
+						   gtk_source_view_get_tab_width (view));
+
+	gtk_source_print_compositor_set_wrap_mode (compositor,
+						   gtk_text_view_get_wrap_mode (GTK_TEXT_VIEW (view)));
+
+	gtk_source_print_compositor_set_print_line_numbers (compositor, 1);
+
+	gtk_source_print_compositor_set_body_font_name (compositor,
+							BODY_FONT_NAME);
+
+	/* To test line numbers font != text font */
+	gtk_source_print_compositor_set_line_numbers_font_name (compositor,
+								LINE_NUMBERS_FONT_NAME);
+
+	gtk_source_print_compositor_set_header_format (compositor,
+						       TRUE,
+						       "Printed on %A",
+						       "test-widget",
+						       "%F");
+
+	gtk_source_print_compositor_set_footer_format (compositor,
+						       TRUE,
+						       "%T",
+						       basename,
+						       "Page %N/%Q");
+
+	gtk_source_print_compositor_set_print_header (compositor, TRUE);
+	gtk_source_print_compositor_set_print_footer (compositor, TRUE);
+
+	gtk_source_print_compositor_set_header_font_name (compositor,
+							  HEADER_FONT_NAME);
+
+	gtk_source_print_compositor_set_footer_font_name (compositor,
+							  FOOTER_FONT_NAME);
+#endif	
+	operation = gtk_print_operation_new ();
+			
+	gtk_print_operation_set_job_name (operation, basename);
+	
+	gtk_print_operation_set_show_progress (operation, TRUE);
+
+#ifndef NON_BLOCKING_PAGINATION	
+  	g_signal_connect (G_OBJECT (operation), "begin-print", 
+			  G_CALLBACK (begin_print), compositor);
+#else
+  	g_signal_connect (G_OBJECT (operation), "paginate", 
+			  G_CALLBACK (paginate), compositor);
+#endif
+	g_signal_connect (G_OBJECT (operation), "draw-page", 
+			  G_CALLBACK (draw_page), compositor);
+	g_signal_connect (G_OBJECT (operation), "end-print", 
+			  G_CALLBACK (end_print), compositor);
+
+	gtk_print_operation_run (operation, 
+				 GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG,
+				 NULL, NULL);
+
+	g_object_unref (operation);
+	g_free (basename);
+}
 
 /* View UI callbacks ------------------------------------------------------------------ */
 
@@ -692,7 +1025,7 @@ update_cursor_position (GtkTextBuffer *buffer, gpointer user_data)
 	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
 
 	view = GTK_SOURCE_VIEW (user_data);
-	tabwidth = gtk_source_view_get_tabs_width (view);
+	tabwidth = gtk_source_view_get_tab_width (view);
 	pos_label = GTK_LABEL (g_object_get_data (G_OBJECT (view), "pos_label"));
 
 	gtk_text_buffer_get_iter_at_mark (buffer,
@@ -781,7 +1114,7 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 	view = GTK_SOURCE_VIEW (widget);
 	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
 
-	if (!gtk_source_view_get_show_line_markers (view))
+	if (!gtk_source_view_get_show_line_marks (view))
 		return FALSE;
 
 	/* check that the click was on the left gutter */
@@ -789,15 +1122,14 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 						    GTK_TEXT_WINDOW_LEFT))
 	{
 		gint y_buf;
-		GtkTextIter line_start, line_end;
-		GSList *marker_list, *list_iter;
-		GtkSourceMarker *marker;
-		const gchar *marker_type;
+		GtkTextIter line_start;
+		GSList *mark_list;
+		const gchar *mark_type;
 
 		if (ev->button == 1)
-			marker_type = MARKER_TYPE_1;
+			mark_type = MARK_TYPE_1;
 		else
-			marker_type = MARKER_TYPE_2;
+			mark_type = MARK_TYPE_2;
 
 		gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (view),
 						       GTK_TEXT_WINDOW_LEFT,
@@ -810,42 +1142,25 @@ button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
 					     y_buf,
 					     NULL);
 
-		line_end = line_start;
-		gtk_text_iter_forward_to_line_end (&line_end);
+		/* get the marks already in the line */
+		mark_list = gtk_source_buffer_get_marks_at_line (buffer,
+								 gtk_text_iter_get_line (&line_start),
+								 mark_type);
 
-		/* get the markers already in the line */
-		marker_list = gtk_source_buffer_get_markers_in_region (buffer,
-								       &line_start,
-								       &line_end);
-
-		/* search for the marker corresponding to the button pressed */
-		marker = NULL;
-		for (list_iter = marker_list;
-		     list_iter && !marker;
-		     list_iter = g_slist_next (list_iter))
+		if (mark_list != NULL)
 		{
-			GtkSourceMarker *tmp = list_iter->data;
-			gchar *tmp_type = gtk_source_marker_get_marker_type (tmp);
-
-			if (tmp_type && !strcmp (tmp_type, marker_type))
-			{
-				marker = tmp;
-			}
-			g_free (tmp_type);
-		}
-		g_slist_free (marker_list);
-
-		if (marker)
-		{
-			/* a marker was found, so delete it */
-			gtk_source_buffer_delete_marker (buffer, marker);
+			/* just take the first and delete it */
+			gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (buffer),
+						     GTK_TEXT_MARK (mark_list->data));
 		}
 		else
 		{
-			/* no marker found -> create one */
-			marker = gtk_source_buffer_create_marker (buffer, NULL,
-								  marker_type, &line_start);
+			/* no mark found: create one */
+			gtk_source_buffer_create_mark (buffer, NULL,
+						       mark_type, &line_start);
 		}
+
+		g_slist_free (mark_list);
 	}
 
 	return FALSE;
@@ -891,9 +1206,12 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 				      G_N_ELEMENTS (view_action_entries), view);
 	gtk_action_group_add_toggle_actions (action_group, toggle_entries,
 					     G_N_ELEMENTS (toggle_entries), view);
-	gtk_action_group_add_radio_actions (action_group, radio_entries,
-					    G_N_ELEMENTS (radio_entries),
+	gtk_action_group_add_radio_actions (action_group, tabs_radio_entries,
+					    G_N_ELEMENTS (tabs_radio_entries),
 					    -1, G_CALLBACK (tabs_toggled_cb), view);
+	gtk_action_group_add_radio_actions (action_group, indent_radio_entries,
+					    G_N_ELEMENTS (indent_radio_entries),
+					    -1, G_CALLBACK (indent_toggled_cb), view);
 	gtk_action_group_add_radio_actions (action_group, smart_home_end_entries,
 					    G_N_ELEMENTS (smart_home_end_entries),
 					    -1, G_CALLBACK (smart_home_end_toggled_cb), view);
@@ -945,24 +1263,29 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 	if (from)
 	{
 		gchar *tmp;
+		gint i;
 		GtkAction *action;
 
 		action = gtk_action_group_get_action (action_group, "ShowNumbers");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_show_line_numbers (from));
 
-		action = gtk_action_group_get_action (action_group, "ShowMarkers");
+		action = gtk_action_group_get_action (action_group, "ShowMarks");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
-					      gtk_source_view_get_show_line_markers (from));
+					      gtk_source_view_get_show_line_marks (from));
 
 		action = gtk_action_group_get_action (action_group, "ShowMargin");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
-					      gtk_source_view_get_show_margin (from));
+					      gtk_source_view_get_show_right_margin (from));
 
 		action = gtk_action_group_get_action (action_group, "HlLine");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_highlight_current_line (from));
 
+		action = gtk_action_group_get_action (action_group, "WrapLines");
+		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+					      gtk_text_view_get_wrap_mode (GTK_TEXT_VIEW (from)) != GTK_WRAP_NONE);
+
 		action = gtk_action_group_get_action (action_group, "AutoIndent");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_auto_indent (from));
@@ -972,24 +1295,31 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 			GTK_TOGGLE_ACTION (action),
 			gtk_source_view_get_insert_spaces_instead_of_tabs (from));
 
+		tmp = g_strdup_printf ("TabWidth%d", gtk_source_view_get_tab_width (from));
+		action = gtk_action_group_get_action (action_group, tmp);
+		if (action)
+			gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
+		g_free (tmp);
 
-		tmp = g_strdup_printf ("TabsWidth%d", gtk_source_view_get_tabs_width (from));
+		i = gtk_source_view_get_indent_width (from);
+		tmp = i < 0 ? g_strdup ("IndentWidthUnset") : g_strdup_printf ("IndentWidth%d", i);
 		action = gtk_action_group_get_action (action_group, tmp);
 		if (action)
 			gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
 		g_free (tmp);
 	}
 
-	/* add marker pixbufs */
+	/* add source mark pixbufs */
 	error = NULL;
 	if ((pixbuf = gdk_pixbuf_new_from_file (DATADIR "/pixmaps/apple-green.png", &error)))
 	{
-		gtk_source_view_set_marker_pixbuf (GTK_SOURCE_VIEW (view), MARKER_TYPE_1, pixbuf);
+		gtk_source_view_set_mark_category_pixbuf (GTK_SOURCE_VIEW (view), MARK_TYPE_1, pixbuf);
+		gtk_source_view_set_mark_category_priority (GTK_SOURCE_VIEW (view), MARK_TYPE_1, 1);
 		g_object_unref (pixbuf);
 	}
 	else
 	{
-		g_message ("could not load marker 1 image.  Spurious messages might get triggered: %s",
+		g_message ("could not load source mark 1 image. Spurious messages might get triggered: %s\n",
 		error->message);
 		g_error_free (error);
 	}
@@ -997,12 +1327,13 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 	error = NULL;
 	if ((pixbuf = gdk_pixbuf_new_from_file (DATADIR "/pixmaps/apple-red.png", &error)))
 	{
-		gtk_source_view_set_marker_pixbuf (GTK_SOURCE_VIEW (view), MARKER_TYPE_2, pixbuf);
+		gtk_source_view_set_mark_category_pixbuf (GTK_SOURCE_VIEW (view), MARK_TYPE_2, pixbuf);
+		gtk_source_view_set_mark_category_priority (GTK_SOURCE_VIEW (view), MARK_TYPE_2, 2);
 		g_object_unref (pixbuf);
 	}
 	else
 	{
-		g_message ("could not load marker 2 image.  Spurious messages might get triggered: %s",
+		g_message ("could not load source mark 2 image. Spurious messages might get triggered: %s\n",
 		error->message);
 		g_error_free (error);
 	}
@@ -1049,7 +1380,7 @@ create_main_window (GtkSourceBuffer *buffer)
 	action = gtk_action_group_get_action (action_group, "ShowNumbers");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
 
-	action = gtk_action_group_get_action (action_group, "ShowMarkers");
+	action = gtk_action_group_get_action (action_group, "ShowMarks");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
 
 	action = gtk_action_group_get_action (action_group, "ShowMargin");
@@ -1061,7 +1392,10 @@ create_main_window (GtkSourceBuffer *buffer)
 	action = gtk_action_group_get_action (action_group, "InsertSpaces");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
 
-	action = gtk_action_group_get_action (action_group, "TabsWidth8");
+	action = gtk_action_group_get_action (action_group, "TabWidth8");
+	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
+
+	action = gtk_action_group_get_action (action_group, "IndentWidthUnset");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
 
 	return window;
@@ -1153,11 +1487,13 @@ main (int argc, char *argv[])
 {
 	GtkWidget *window;
 	GtkSourceLanguageManager *lm;
-	GtkSourceStyleManager *sm;
+	GtkSourceStyleSchemeManager *sm;
 	GtkSourceBuffer *buffer;
 
 	gchar *builtin_lang_dirs[] = {TOP_SRCDIR "/gtksourceview/language-specs", NULL};
+	gchar *builtin_sm_dirs[] = {TOP_SRCDIR "/gtksourceview/language-specs", NULL};
 	gchar **lang_dirs;
+	const gchar * const * schemes;
 	gboolean use_default_paths = FALSE;
 
 	gchar *style_scheme_id = NULL;
@@ -1190,11 +1526,47 @@ main (int argc, char *argv[])
 	lang_dirs = use_default_paths ? NULL : builtin_lang_dirs;
 	lm = gtk_source_language_manager_get_default ();
 	gtk_source_language_manager_set_search_path (lm, lang_dirs);
-	sm = gtk_source_style_manager_get_default ();
-	gtk_source_style_manager_set_search_path (sm, lang_dirs);
+
+	lang_dirs = use_default_paths ? NULL : builtin_sm_dirs;
+
+	sm = gtk_source_style_scheme_manager_get_default ();
+	gtk_source_style_scheme_manager_set_search_path (sm, lang_dirs);
+
+	if (!use_default_paths)
+		gtk_source_style_scheme_manager_append_search_path (sm, TOP_SRCDIR "/tests/test-scheme.xml");
+
+	schemes = gtk_source_style_scheme_manager_get_scheme_ids (sm);
+	g_print ("Available style schemes:\n");
+	while (*schemes != NULL)
+	{
+		const gchar* const *authors;
+		gchar *authors_str = NULL;
+
+		style_scheme = gtk_source_style_scheme_manager_get_scheme (sm, *schemes);
+
+		authors = gtk_source_style_scheme_get_authors (style_scheme);
+		if (authors != NULL)
+			authors_str = g_strjoinv (", ", (gchar **)authors);
+
+		g_print (" - [%s] %s: %s\n",
+			 gtk_source_style_scheme_get_id (style_scheme),
+			 gtk_source_style_scheme_get_name (style_scheme),
+			 gtk_source_style_scheme_get_description (style_scheme) ?
+				gtk_source_style_scheme_get_description (style_scheme) : "");
+
+		if (authors_str) {
+			g_print ("   by %s\n",  authors_str);
+			g_free (authors_str);
+		}
+
+		++schemes;
+	}
+	g_print("\n");
 
 	if (style_scheme_id != NULL)
-		style_scheme = gtk_source_style_manager_get_scheme (sm, style_scheme_id);
+		style_scheme = gtk_source_style_scheme_manager_get_scheme (sm, style_scheme_id);
+	else
+		style_scheme = NULL;
 
 	/* create buffer */
 	buffer = gtk_source_buffer_new (NULL);
